""" flywheel """
from dynamo3 import (CheckFailed, ConditionalCheckFailedException, Binary,
                     STRING, NUMBER, BINARY, STRING_SET, NUMBER_SET,
                     BINARY_SET, Limit)

from .fields import Field, Composite, GlobalIndex
from .fields.types import set_, TypeDefinition
from .models import Model
from .engine import Engine
from .query import DuplicateEntityException, EntityNotFoundException

__version__ = '0.4.7'



""" Unit and system tests for flywheel """
import six
from flywheel.engine import Engine


try:
    import unittest2 as unittest  # pylint: disable=F0401
except ImportError:
    import unittest


class DynamoSystemTest(unittest.TestCase):

    """ Base class for tests that need an :class:`~flywheel.engine.Engine` """
    dynamo = None
    models = []

    @classmethod
    def setUpClass(cls):
        super(DynamoSystemTest, cls).setUpClass()
        cls.engine = Engine(cls.dynamo, ['test'])
        cls.engine.register(*cls.models)
        cls.engine.create_schema()

    @classmethod
    def tearDownClass(cls):
        super(DynamoSystemTest, cls).tearDownClass()
        cls.engine.delete_schema()

    def tearDown(self):
        super(DynamoSystemTest, self).tearDown()
        for model in six.itervalues(self.engine.models):
            self.engine.scan(model).delete()



""" Index definitions """
from dynamo3 import DynamoKey, Throughput
from dynamo3 import GlobalIndex as GIndex


class GlobalIndex(object):

    """
    A global index for DynamoDB

    Parameters
    ----------
    name : str
        The name of the index
    hash_key : str
        The name of the field that is the hash key for the index
    range_key : str, optional
        The name of the field that is the range key for the index
    throughput : dict, optional
        The read/write throughput of this global index. Used when creating a
        table. Dict has a 'read' and a 'write' key. (Default 5, 5)

    """

    def __init__(self, name, hash_key, range_key=None):
        self.name = name
        self.hash_key = hash_key
        self.range_key = range_key
        self._throughput = Throughput()
        self.ddb_index = GIndex.all
        self.kwargs = {}

    @classmethod
    def all(cls, name, hash_key, range_key=None):
        """ Project all attributes into the index """
        return cls(name, hash_key, range_key)

    @classmethod
    def keys(cls, name, hash_key, range_key=None):
        """ Project key attributes into the index """
        index = cls(name, hash_key, range_key)
        index.ddb_index = GIndex.keys
        return index

    @classmethod
    def include(cls, name, hash_key, range_key=None, includes=None):
        """ Select which attributes to project into the index """
        includes = includes or []
        index = cls(name, hash_key, range_key)
        index.ddb_index = GIndex.include
        index.kwargs['includes'] = includes
        return index

    def get_ddb_index(self, fields):
        """ Get the dynamo index class for this GlobalIndex """
        hash_key = DynamoKey(self.hash_key,
                             data_type=fields[self.hash_key].ddb_data_type)
        range_key = None
        if self.range_key is not None:
            range_key = DynamoKey(self.range_key,
                                  data_type=fields[self.range_key].ddb_data_type)
        index = self.ddb_index(self.name, hash_key, range_key,
                               throughput=self._throughput, **self.kwargs)
        return index

    def throughput(self, read=5, write=5):
        """
        Set the index throughput

        Parameters
        ----------
        read : int, optional
            Amount of read throughput (default 5)
        write : int, optional
            Amount of write throughput (default 5)

        Notes
        -----
        This is meant to be used as a chain::

            class MyModel(Model):
                __metadata__ = {
                    'global_indexes': [
                        GlobalIndex('myindex', 'hkey', 'rkey').throughput(5, 2)
                    ]
                }

        """
        self._throughput = Throughput(read, write)
        return self

    def __contains__(self, field):
        return field == self.hash_key or field == self.range_key

    def __iter__(self):
        yield self.hash_key
        if self.range_key is not None:
            yield self.range_key

    def __repr__(self):
        if self.range_key is None:
            return "GlobalIndex('%s', '%s')" % (self.name, self.hash_key)
        else:
            return "GlobalIndex('%s', '%s', '%s')" % (self.name, self.hash_key,
                                                      self.range_key)

    def __str__(self):
        return repr(self)



""" Query engine """
import itertools

import logging
import six
from collections import defaultdict
from dynamo3 import DynamoDBConnection, CheckFailed, ItemUpdate, ALL_NEW

from .fields import Field
from .models import Model, SetDelta
from .query import Query, Scan


LOG = logging.getLogger(__name__)


class Engine(object):

    """
    Query engine for models

    Parameters
    ----------
    dynamo : :class:`dynamodb3.DynamoDBConnection`, optional
    namespace : list or str, optional
        String prefix or list of component parts of a prefix for models. All
        table names will be prefixed by this string or strings (joined by '-').
    default_conflict : {'update', 'overwrite', 'raise'}, optional
        Default setting for delete(), save(), and sync() (default 'update')

    Notes
    -----
    The engine is used to save, sync, delete, and query DynamoDB. Here is a
    basic example of saving items:

    .. code-block:: python

        item1 = MyModel()
        engine.save(item1)
        item1.foobar = 'baz'
        item2 = MyModel()
        engine.save([item1, item2], overwrite=True)

    You can also use the engine to query tables:

    .. code-block:: python

        user = engine.query(User).filter(User.id == 'abcdef).first()

        # calling engine() is a shortcut for engine.query()
        user = engine(User).filter(User.id == 'abcdef).first()

        d_users = engine(User).filter(User.school == 'MIT',
                                      User.name.beginswith_('D')).all()

        # You can pass in equality constraints as keyword args
        user = engine(User).filter(id='abcdef').first()

    Scans are like queries, except that they don't use an index. Scans iterate
    over the ENTIRE TABLE so they are REALLY SLOW. Scans have access to
    additional filter conditions such as "contains" and "in".

    .. code-block:: python

        # This is suuuuuper slow!
        user = engine.scan(User).filter(id='abcdef').first()

        # If you're doing an extremely large scan, you should tell it to return
        # a generator
        all_users = engine.scan(User).gen()

        # to filter a field not specified in the model declaration:
        prince = engine.scan(User).filter(User.field_('bio').beginswith_(
                   'Now this is a story all about how')).first()

    """

    def __init__(self, dynamo=None, namespace=(), default_conflict='update'):
        if dynamo is not None and not isinstance(dynamo, DynamoDBConnection):
            raise ValueError("'dynamo' needs to be an instance of "
                             "dynamo3.DynamoDBConnection (are you using a "
                             "boto connection?)")
        self.dynamo = dynamo
        self.models = {}
        self.namespace = namespace
        self._default_conflict = None
        self.default_conflict = default_conflict

    @property
    def default_conflict(self):
        """
        Get the default_conflict value

        Notes
        -----
        The ``default_conflict`` setting configures the default behavior of
        :meth:`~.Engine.save`, :meth:`~.Engine.sync`, and
        :meth:`~.Engine.delete`. Below is an explanation of the different
        values of ``default_conflict``.

        +------------------+--------+-------------------------+
        | default_conflict | method | default                 |
        +==================+========+=========================+
        |**'update'**      |        |                         |
        +------------------+--------+-------------------------+
        |                  | save   | overwrite=True          |
        +------------------+--------+-------------------------+
        |                  | sync   | raise_on_conflict=True  |
        +------------------+--------+-------------------------+
        |                  | delete | raise_on_conflict=False |
        +------------------+--------+-------------------------+
        |**'overwrite'**   |        |                         |
        +------------------+--------+-------------------------+
        |                  | save   | overwrite=True          |
        +------------------+--------+-------------------------+
        |                  | sync   | raise_on_conflict=False |
        +------------------+--------+-------------------------+
        |                  | delete | raise_on_conflict=False |
        +------------------+--------+-------------------------+
        |**'raise'**       |        |                         |
        +------------------+--------+-------------------------+
        |                  | save   | overwrite=False         |
        +------------------+--------+-------------------------+
        |                  | sync   | raise_on_conflict=True  |
        +------------------+--------+-------------------------+
        |                  | delete | raise_on_conflict=True  |
        +------------------+--------+-------------------------+

        """
        return self._default_conflict

    @default_conflict.setter
    def default_conflict(self, default_conflict):
        """ Protected setter for default_conflict """
        if default_conflict not in ('update', 'overwrite', 'raise'):
            raise ValueError("Unrecognized value '%s' for default_conflict" %
                             default_conflict)
        self._default_conflict = default_conflict

    def connect_to_region(self, region, **kwargs):
        """ Connect to an AWS region """
        self.dynamo = DynamoDBConnection.connect(region, **kwargs)

    def connect_to_host(self, **kwargs):
        """ Connect to a specific host """
        self.dynamo = DynamoDBConnection.connect_to_host(**kwargs)

    def connect(self, *args, **kwargs):
        """ Connect to a specific host """
        self.dynamo = DynamoDBConnection.connect(*args, **kwargs)

    def register(self, *models):
        """
        Register one or more models with the engine

        Registering is required for schema creation or deletion

        """
        for model in models:
            if model.meta_.name in self.models:
                raise ValueError("Model name '%s' already registered!" %
                                 model.meta_.name)
            self.models[model.meta_.name] = model

    def create_schema(self, test=False, throughput=None):
        """
        Create the DynamoDB tables required by the registered models

        Parameters
        ----------
        test : bool, optional
            If True, perform a dry run (default False)
        throughput : dict, optional
            If provided, override the throughputs of the Models when creating
            them. Details below.

        Returns
        -------
        names : list
            List of table names that were created

        Examples
        --------
        The ``throughput`` argument is a mapping of table names to their
        throughputs. The throughput is a dict with a 'read' and 'write' value.
        It may also include the names of global indexes that map to their own
        dicts with a 'read' and 'write' value.

        .. code-block:: python

            engine.create_schema(throughput={
                'table1': {
                    'read': 4,
                    'write': 10,
                    'gindex-1': {
                        'read': 6,
                        'write': 3,
                    }
                }
            })

        """
        throughput = throughput or {}
        tablenames = set(self.dynamo.list_tables())
        changed = []
        for model in six.itervalues(self.models):
            result = model.meta_.create_dynamo_schema(
                self.dynamo, tablenames, test=test, wait=True,
                throughput=throughput.get(model.meta_.ddb_tablename()),
                namespace=self.namespace)
            if result:
                changed.append(result)
        return changed

    def delete_schema(self, test=False):
        """
        Drop the DynamoDB tables for all registered models

        Parameters
        ----------
        test : bool, optional
            If True, perform a dry run (default False)

        Returns
        -------
        names : list
            List of table names that were deleted

        """
        tablenames = set(self.dynamo.list_tables())
        changed = []
        for model in six.itervalues(self.models):
            result = model.meta_.delete_dynamo_schema(self.dynamo, tablenames,
                                                      test=test, wait=True,
                                                      namespace=self.namespace)
            changed.append(result)
        return changed

    def get_schema(self):
        """ Get the schema for the registered models """
        schema = []
        for model in six.itervalues(self.models):
            schema.append(model.meta_.ddb_tablename(self.namespace))
        return schema

    def __call__(self, model):
        """ Shortcut for query """
        return self.query(model)

    def query(self, model):
        """
        Create a table query for a specific model

        Returns
        -------
        query : :class:`.Query`

        """
        return Query(self, model)

    def scan(self, model):
        """
        Create a table scan for a specific model

        Returns
        -------
        scan : :class:`.Scan`

        """
        return Scan(self, model)

    def get(self, model, pkeys=None, consistent=False, **kwargs):
        """
        Fetch one or more items from dynamo from the primary keys

        Parameters
        ----------
        model : :class:`~flywheel.models.Model`
        pkeys : list, optional
            List of primary key dicts
        consistent : bool, optional
            Perform a consistent read from dynamo (default False)
        **kwargs : dict
            If pkeys is None, fetch only a single item and use kwargs as the
            primary key dict.

        Returns
        -------
        items : list or object
            If pkeys is a list of key dicts, this will be a list of items. If
            pkeys is None and **kwargs is used, this will be a single object.

        Notes
        -----
        If the model being fetched has no range key, you may use strings
        instead of primary key dicts. ex:

        .. code-block:: python

            >>> class Item(Model):
            ...     id = Field(hash_key=True)
            ...
            >>> items = engine.get(Item, ['abc', 'def', '123', '456'])

        """
        if pkeys is not None:
            if len(pkeys) == 0:
                return []
            keys = [model.meta_.pk_dict(scope=key) for key in pkeys]
        else:
            keys = [model.meta_.pk_dict(scope=kwargs)]

        raw_items = self.dynamo.batch_get(
            model.meta_.ddb_tablename(self.namespace), keys,
            consistent=consistent)
        items = [model.ddb_load_(self, raw_item) for raw_item in raw_items]
        if pkeys is not None:
            return items
        if len(items) > 0:
            return items[0]
        return None

    def delete_key(self, model, pkeys=None, **kwargs):
        """
        Delete one or more items from dynamo as specified by primary keys

        Parameters
        ----------
        model : :class:`~flywheel.models.Model`
        pkeys : list, optional
            List of primary key dicts
        **kwargs : dict
            If pkeys is None, delete only a single item and use kwargs as the
            primary key dict

        Returns
        -------
        count : int
            The number of deleted items

        Notes
        -----
        If the model being deleted has no range key, you may use strings
        instead of primary key dicts. ex:

        .. code-block:: python

            >>> class Item(Model):
            ...     id = Field(hash_key=True)
            ...
            >>> items = engine.delete_key(Item, ['abc', 'def', '123', '456'])

        """
        if pkeys is not None:
            keys = pkeys
        else:
            keys = [kwargs]

        count = 0
        tablename = model.meta_.ddb_tablename(self.namespace)
        with self.dynamo.batch_write(tablename) as batch:
            for key in keys:
                pkey = model.meta_.pk_dict(scope=key)
                batch.delete(pkey)
                count += 1
        return count

    # Alias because it makes sense
    delete_keys = delete_key

    def delete(self, items, raise_on_conflict=None):
        """
        Delete items from dynamo

        Parameters
        ----------
        items : list or :class:`~flywheel.model.Model`
            List of :class:`~flywheel.models.Model` objects to delete
        raise_on_conflict : bool, optional
            If True, raise exception if the object was changed concurrently in
            the database (default set by :attr:`.default_conflict`)

        Raises
        ------
        exc : :class:`dynamo3.ConditionalCheckFailedException`
            If overwrite is False and an item already exists in the database

        Notes
        -----
        Due to the structure of the AWS API, deleting with
        raise_on_conflict=False is much faster because the requests can be
        batched.

        """
        if raise_on_conflict is None:
            raise_on_conflict = self.default_conflict == 'raise'
        if isinstance(items, Model):
            items = [items]
        if not items:
            return
        tables = defaultdict(list)
        for item in items:
            tables[item.meta_.ddb_tablename(self.namespace)].append(item)

        count = 0
        for tablename, items in six.iteritems(tables):
            if raise_on_conflict:
                for item in items:
                    expected = item.construct_ddb_expects_()
                    count += 1
                    self.dynamo.delete_item(tablename, item.pk_dict_,
                                            **expected)
            else:
                with self.dynamo.batch_write(tablename) as batch:
                    for item in items:
                        if isinstance(item, Model):
                            keys = item.pk_dict_
                        else:
                            keys = item
                        count += 1
                        batch.delete(keys)
        return count

    def save(self, items, overwrite=None):
        """
        Save models to dynamo

        Parameters
        ----------
        items : list or :class:`~flywheel.models.Model`
        overwrite : bool, optional
            If False, raise exception if item already exists (default set by
            :attr:`.default_conflict`)

        Raises
        ------
        exc : :class:`dynamo3.ConditionalCheckFailedException`
            If overwrite is False and an item already exists in the database

        Notes
        -----
        Overwrite will replace the *entire* item with the new one, not just
        different fields. After calling save(overwrite=True) you are guaranteed
        that the item in the database is exactly the item you saved.

        Due to the structure of the AWS API, saving with overwrite=True is much
        faster because the requests can be batched.

        """
        if overwrite is None:
            overwrite = self.default_conflict in ('update', 'overwrite')
        if isinstance(items, Model):
            items = [items]
        if not items:
            return
        tables = defaultdict(list)
        for item in items:
            tables[item.meta_.ddb_tablename(self.namespace)].append(item)
        for tablename, items in six.iteritems(tables):
            if overwrite:
                with self.dynamo.batch_write(tablename) as batch:
                    for item in items:
                        item.pre_save_(self)
                        batch.put(item.ddb_dump_())
                        item.post_save_()
            else:
                for item in items:
                    expected = dict(((k + '__null', True) for k in
                                     item.pk_dict_))
                    item.pre_save_(self)
                    self.dynamo.put_item(tablename, item.ddb_dump_(),
                                         **expected)
                    item.post_save_()

    def refresh(self, items, consistent=False):
        """
        Overwrite model data with freshest from database

        Parameters
        ----------
        items : list or :class:`~flywheel.models.Model`
            Models to sync
        consistent : bool, optional
            If True, force a consistent read from the db. (default False)

        """
        if isinstance(items, Model):
            items = [items]
        if not items:
            return

        tables = defaultdict(list)
        model_map = defaultdict(dict)
        for item in items:
            tablename = item.meta_.ddb_tablename(self.namespace)
            tables[tablename].append(item)
            model_map[tablename][item.pk_tuple_] = item

        for tablename, items in six.iteritems(tables):
            keys = [item.pk_dict_ for item in items]
            results = self.dynamo.batch_get(tablename, keys,
                                            consistent=consistent)
            meta = items[0].meta_
            for result in results:
                pkey = meta.pk_tuple(None, result, ddb_load=True)
                item = model_map[tablename].get(pkey)
                if item is None:
                    LOG.error("Refresh error: Cannot match primary key %r to "
                              "a model", pkey)
                    continue
                with item.loading_(self):
                    for key, val in six.iteritems(result):
                        item.set_ddb_val_(key, val)

    def sync(self, items, raise_on_conflict=None, consistent=False, constraints=None):
        """
        Sync model changes back to database

        This will push any updates to the database, and ensure that all of the
        synced items have the most up-to-date data.

        Parameters
        ----------
        items : list or :class:`~flywheel.models.Model`
            Models to sync
        raise_on_conflict : bool, optional
            If True, raise exception if any of the fields that are being
            updated were concurrently changed in the database (default set by
            :attr:`.default_conflict`)
        consistent : bool, optional
            If True, force a consistent read from the db. This will only take
            effect if the sync is only performing a read. (default False)
        constraints : list, optional
            List of more complex constraints that must pass for the update to
            complete. Must be used with raise_on_conflict=True. Format is the
            same as query filters (e.g. Model.fieldname > 5)

        Raises
        ------
        exc : :class:`dynamo3.CheckFailed`
            If raise_on_conflict=True and the model changed underneath us

        """
        if raise_on_conflict is None:
            raise_on_conflict = self.default_conflict in ('update', 'raise')
        if constraints is not None and not raise_on_conflict:
            raise ValueError("Cannot pass constraints to sync() when raise_on_conflict is False")
        if isinstance(items, Model):
            items = [items]
        refresh_models = []
        for item in items:
            # Look for any mutable fields (e.g. sets) that have changed
            for name in item.keys_():
                if name in item.__dirty__ or name in item.__incrs__:
                    continue
                field = item.meta_.fields.get(name)
                if field is None:
                    value = item.get_(name)
                    if Field.is_overflow_mutable(value):
                        if value != item.cached_(name):
                            item.__dirty__.add(name)
                elif field.is_mutable:
                    cached_var = item.cached_(name)
                    if field.resolve(item) != cached_var:
                        for related in item.meta_.related_fields[name]:
                            item.__dirty__.add(related)

            if not item.__dirty__ and not item.__incrs__:
                refresh_models.append(item)
                continue
            fields = item.__dirty__
            item.pre_save_(self)

            # If the model has changed any field that is part of a composite
            # field, FORCE the sync to raise on conflict. This prevents the
            # composite key from potentially getting into an inconsistent state
            _raise_on_conflict = raise_on_conflict
            for name in itertools.chain(item.__incrs__, fields):
                for related_name in item.meta_.related_fields.get(name, []):
                    field = item.meta_.fields[related_name]
                    if field.composite:
                        _raise_on_conflict = True
                        break

            keywords = {}
            constrained_fields = set()
            if _raise_on_conflict and constraints is not None:
                for constraint in constraints:
                    constrained_fields.update(constraint.eq_fields.keys())
                    constrained_fields.update(constraint.fields.keys())
                    keywords.update(constraint.scan_kwargs())

            updates = []
            # Set dynamo keys
            for name in fields:
                field = item.meta_.fields.get(name)
                value = getattr(item, name)
                kwargs = {}
                if _raise_on_conflict and name not in constrained_fields:
                    kwargs = {'eq': item.ddb_dump_cached_(name)}
                update = ItemUpdate.put(name, item.ddb_dump_field_(name),
                                        **kwargs)
                updates.append(update)

            # Atomic increment fields
            for name, value in six.iteritems(item.__incrs__):
                kwargs = {}
                # We don't need to ddb_dump because we know they're all native
                if isinstance(value, SetDelta):
                    update = ItemUpdate(value.action, name, value.values)
                else:
                    update = ItemUpdate.add(name, value)
                updates.append(update)

            # Perform sync
            ret = self.dynamo.update_item(
                item.meta_.ddb_tablename(self.namespace), item.pk_dict_,
                updates, returns=ALL_NEW, **keywords)

            # Load updated data back into object
            with item.loading_(self):
                for key, val in six.iteritems(ret):
                    item.set_ddb_val_(key, val)

            item.post_save_()

        # Handle items that didn't have any fields to update

        # If the item isn't known to exist in the db, try to save it first
        for item in refresh_models:
            if not item.persisted_:
                try:
                    self.save(item, overwrite=False)
                except CheckFailed:
                    pass
        # Refresh item data
        self.refresh(refresh_models, consistent=consistent)



""" Model metadata and metaclass objects """
import six
import time

import inspect
from dynamo3 import DynamoKey, Throughput
from collections import defaultdict

from .fields import Field
from .fields.conditions import FILTER_ONLY


class ValidationError(Exception):

    """ Model inconsistency """
    pass


class Ordering(object):

    """
    A way that the models are ordered

    This will be a combination of a hash key and a range key. It may be the
    primary key, a local secondary index, or a global secondary index.

    """

    def __init__(self, meta, hash_key, range_key=None, index_name=None):
        self.meta = meta
        self.hash_key = hash_key
        self.range_key = range_key
        self.index_name = index_name

    def query_kwargs(self, eq_fields, fields):
        """ Get the query and filter kwargs for querying against this index """
        kwargs = {'%s__eq' % self.hash_key.name:
                  self.hash_key.resolve(scope=eq_fields)}
        if self.index_name is not None:
            kwargs['index'] = self.index_name
        remaining = set(eq_fields)
        remaining = remaining.union(fields)
        remaining -= self.hash_key.can_resolve(eq_fields)
        if self.range_key is not None:
            eq_range_fields = self.range_key.can_resolve(eq_fields)
            if eq_range_fields:
                remaining -= eq_range_fields
                key = '%s__eq' % self.range_key.name
                val = self.range_key.resolve(scope=eq_fields)
                kwargs[key] = val
            else:
                for field in self.range_key.can_resolve(fields):
                    (op, val) = fields[field]
                    if op in FILTER_ONLY:
                        continue
                    kwargs['%s__%s' % (field, op)] = val
                    remaining.remove(field)

        # Find the additional filter arguments
        filter_fields = {}
        for key in remaining:
            if key in eq_fields:
                filter_fields['%s__eq' % key] = eq_fields[key]
            else:
                op, val = fields[key]
                filter_fields['%s__%s' % (key, op)] = val
        kwargs['filter'] = filter_fields

        return kwargs

    def pk_dict(self, obj=None, scope=None, ddb_dump=False):
        """ Get the dynamo primary key dict for this ordering """
        # If we can unambiguously tell that a single string defines the primary
        # key, allow scope to be a single string
        if (obj is None and isinstance(scope, six.string_types) and
                self.range_key is None):
            scope = {self.hash_key.name: scope}

        hk = self.hash_key.resolve(obj, scope)
        if ddb_dump:
            hk = self.hash_key.ddb_dump(hk)
        key_dict = {self.hash_key.name: hk}
        if self.range_key is not None:
            rk = self.range_key.resolve(obj, scope)
            if ddb_dump:
                rk = self.range_key.ddb_dump(rk)
            key_dict[self.range_key.name] = rk
        return key_dict

    def __repr__(self):
        if self.range_key is None:
            return "Ordering(%s, None, %s)" % (self.hash_key.name,
                                               self.index_name)
        else:
            return "Ordering(%s, %s, %s)" % (self.hash_key.name,
                                             self.range_key.name,
                                             self.index_name)


def merge_metadata(cls):
    """
    Merge all the __metadata__ dicts in a class's hierarchy

    keys that do not begin with '_' will be inherited.

    keys that begin with '_' will only apply to the object that defines them.

    """
    cls_meta = cls.__dict__.get('__metadata__', {})
    meta = {}
    for base in cls.__bases__:
        meta.update(getattr(base, '__metadata__', {}))
    # Don't merge any keys that start with '_'
    for key in list(meta.keys()):
        if key.startswith('_'):
            del meta[key]
    meta.update(cls_meta)
    return meta


class ModelMetaclass(type):

    """
    Metaclass for Model objects

    Merges model metadata, sets the ``meta_`` attribute, and performs
    validation checks.

    """
    def __new__(mcs, name, bases, dct):
        cls = super(ModelMetaclass, mcs).__new__(mcs, name, bases, dct)

        cls.__metadata__ = merge_metadata(cls)
        if hasattr(cls, '__on_create__'):
            cls.__on_create__()

        if hasattr(cls, '__metadata_class__'):
            cls.meta_ = cls.__metadata_class__(cls)
            cls.meta_.post_create()
            cls.meta_.validate_model()
            cls.meta_.post_validate()

        if hasattr(cls, '__after_create__'):
            cls.__after_create__()

        return cls

    def __init__(cls, name, bases, dct):
        super(ModelMetaclass, cls).__init__(name, bases, dct)


class ModelMetadata(object):

    """
    Container for model metadata

    Parameters
    ----------
    model : :class:`.Model`

    Attributes
    ----------
    name : str
        The unique name of the model. This is set by the '_name' field in
        __metadata__. Defaults to the name of the model class.
    abstract : bool
        If a model is abstract then it has no table in Dynamo
    global_indexes : list
        List of global indexes (hash_key, [range_key]) pairs.
    related_fields : dict
        Mapping of field names to set of fields that change when that field
        changes (usually just that field name, but can be more if composite
        fields use it)
    orderings : list
        List of :class:`.Ordering`
    throughput : dict
        Mapping of 'read' and 'write' to the table throughput (default 5, 5)

    """
    __order_class__ = Ordering

    def __init__(self, model):
        self.model = model
        self._name = model.__name__
        self.global_indexes = []
        self.orderings = []
        self.throughput = Throughput()
        self._abstract = False
        self.__dict__.update(model.__metadata__)
        # Allow throughput to be specified as read/write in a dict
        # pylint: disable=E1134
        if isinstance(self.throughput, dict):
            self.throughput = Throughput(**self.throughput)
        # pylint: enable=E1134
        self.name = self._name
        self.fields = {}
        self.hash_key = None
        self.range_key = None
        self.related_fields = defaultdict(set)
        self.all_global_indexes = set()
        for gindex in self.global_indexes:
            self.all_global_indexes.add(gindex.hash_key)
            if gindex.range_key is not None:
                self.all_global_indexes.add(gindex.range_key)

    def post_create(self):
        """ Create the orderings """
        for name, member in inspect.getmembers(self.model):
            if isinstance(member, Field):
                if name.startswith('__') or name.endswith('_'):
                    raise ValidationError("Field '%s' cannot begin with '__' "
                                          "or end with '_'" % name)
                self.fields[name] = member
                member.name = name
                member.model = self.model
                if member.hash_key:
                    self.hash_key = member
                elif member.range_key:
                    self.range_key = member

        self.orderings.append(self.__order_class__(self, self.hash_key,
                                                   self.range_key))
        for field in six.itervalues(self.fields):
            if field.index:
                order = self.__order_class__(self, self.hash_key, field,
                                             field.index_name)
                self.orderings.append(order)

        for index in self.global_indexes:
            for key in index:
                if key not in self.fields:
                    raise ValidationError("Model %s global index %s "
                                          "references unknown field '%s'" %
                                          (self.name, index.name, key))
            if index.range_key:
                range_key = self.fields[index.range_key]
            else:
                range_key = None
            self.orderings.append(self.__order_class__(self,
                                                       self.fields[
                                                           index.hash_key],
                                                       range_key, index.name))

    def post_validate(self):
        """ Build the dict of related fields """
        def update_related(field, name):
            """ Recursively add a field to related """
            for f in field.subfields:
                self.related_fields[f].add(name)
                subfield = self.fields[f]
                if subfield.composite:
                    update_related(subfield, name)

        for field in six.itervalues(self.fields):
            self.related_fields[field.name].add(field.name)
            if field.composite:
                update_related(field, field.name)

    def get_ordering_from_fields(self, eq_fields, fields):
        """
        Get a unique ordering from constraint fields.

        This does a best-effort guess of which index is being queried. It
        prioritizes indexes that have a constraint on the range key. It
        prioritizes the primary key over local and global indexes.

        Parameters
        ----------
        eq_fields : list
            List of field names that are constrained with '='.
        fields : list
            List of field names that are constrained with inequality operators
            ('>', '<', 'beginswith', etc)

        Returns
        -------
        ordering : :class:`.Ordering`

        Raises
        ------
        exc : :class:`TypeError`
            If more than one possible Ordering is found

        """
        index_satisfied_orderings = []
        other_orderings = []
        eq_field_set = set(eq_fields)
        for order in self.orderings:
            needed = order.hash_key.can_resolve(eq_fields)
            # hash key could not be satisfied
            if len(needed) == 0:
                continue

            remaining = eq_field_set - needed

            if order.range_key is not None:
                if order.range_key.can_resolve(remaining) or order.range_key.can_resolve(fields):
                    index_satisfied_orderings.append(order)
                    continue
            other_orderings.append(order)

        def get_best_ordering(orderings):
            """ Find the best choice in a list of orderings. """
            if len(orderings) == 1:
                return orderings[0]
            else:
                for order in orderings:
                    if order.index_name is None:
                        return order
                raise ValueError("More than one ordering found: %s" % orderings)

        if index_satisfied_orderings:
            return get_best_ordering(index_satisfied_orderings)
        elif other_orderings:
            return get_best_ordering(other_orderings)
        else:
            return None

    def get_ordering_from_index(self, index):
        """ Get the ordering with matching index name """
        for order in self.orderings:
            if order.index_name == index:
                return order
        raise ValueError("Cannot find ordering with index name '%s'" % index)

    def rk(self, obj=None, scope=None):
        """ Construct the range key value """
        if self.range_key is not None:
            return self.range_key.resolve(obj, scope)

    def hk(self, obj=None, scope=None):
        """ Construct the primary key value """
        return self.hash_key.resolve(obj, scope)

    def pk_tuple(self, obj=None, scope=None, ddb_dump=False, ddb_load=False):
        """ Get a tuple that represents the primary key for an item """
        hk = self.hk(obj, scope)
        if ddb_dump:
            hk = self.hash_key.ddb_dump(hk)
        elif ddb_load:
            hk = self.hash_key.ddb_load(hk)
        if self.range_key is None:
            return (hk,)
        rk = self.rk(obj, scope)
        if ddb_dump:
            rk = self.range_key.ddb_dump(rk)
        elif ddb_load:
            rk = self.range_key.ddb_load(rk)
        return (hk, rk)

    def pk_dict(self, obj=None, scope=None, ddb_dump=False):
        """ Get the dynamo primary key dict for an item """
        return self.index_pk_dict(None, obj, scope, ddb_dump)

    def index_pk_dict(self, index_name, obj=None, scope=None, ddb_dump=False):
        """ Get the primary key dict for an index (includes the table key) """
        # Get the 'table' index, which is the hash & range key
        table_order = self.get_ordering_from_index(None)
        pk = table_order.pk_dict(obj, scope, ddb_dump)
        if index_name is not None:
            ordering = self.get_ordering_from_index(index_name)
            pk.update(ordering.pk_dict(obj, scope, ddb_dump))
        return pk

    @property
    def abstract(self):
        """ Getter for abstract """
        return self._abstract

    def ddb_tablename(self, namespace=()):
        """
        The name of the DynamoDB table

        Parameters
        ----------
        namespace : list or str, optional
            String prefix or list of component parts of a prefix for the table
            name.  The prefix will be this string or strings (joined by '-').

        """
        if self.abstract:
            return None
        elif isinstance(namespace, six.string_types):
            return namespace + self.name
        else:
            return '-'.join(tuple(namespace) + (self.name,))

    def validate_model(self):
        """ Perform validation checks on the model declaration """
        if self.abstract or self.model.__dict__.get('__abstract__'):
            return
        hash_keys = [f for f in self.fields.values() if f.hash_key]
        range_keys = [f for f in self.fields.values() if f.range_key]
        indexes = [f for f in self.fields.values() if f.index]
        name = self.name
        if len(hash_keys) != 1:
            raise ValidationError("Model %s must have exactly one hash key" %
                                  name)
        if len(range_keys) > 1:
            raise ValidationError("Model %s can't have more than one range key"
                                  % name)
        if len(indexes) > 5:
            raise ValidationError("Model %s can't have more than 5 "
                                  "local indexes" % name)
        if len(indexes) > 0 and len(range_keys) == 0:
            raise ValidationError("Model %s can't set indexes without a "
                                  "range key" % name)
        if len(self.global_indexes) > 5:
            raise ValidationError("Model %s can't have more than 5 "
                                  "global indexes" % name)
        for field in self.fields.values():
            if field.composite:
                for f in field.subfields:
                    if f not in self.fields:
                        raise ValidationError("Model %s key %s references "
                                              "unknown field '%s'" %
                                              (name, field.name, f))
                    if f == field.name:
                        raise ValidationError("Model %s key %s cannot contain "
                                              "itself" % (name, field.name))

    def create_dynamo_schema(self, connection, tablenames=None, test=False,
                             wait=False, throughput=None, namespace=()):
        """
        Create all Dynamo tables for this model

        Parameters
        ----------
        connection : :class:`~dynamo3.DynamoDBConnection`
        tablenames : list, optional
            List of tables that already exist. Will call 'describe' if not
            provided.
        test : bool, optional
            If True, don't actually create the table (default False)
        wait : bool, optional
            If True, block until table has been created (default False)
        throughput : dict, optional
            The throughput of the table and global indexes. Has the keys 'read'
            and 'write'. To specify throughput for global indexes, add the name
            of the index as a key and another 'read', 'write' dict as the
            value.
        namespace : str or tuple, optional
            The namespace of the table

        Returns
        -------
        table : str
            Table name that was created, or None if nothing created

        """
        if self.abstract:
            return None
        if tablenames is None:
            tablenames = set(connection.list_tables())
        tablename = self.ddb_tablename(namespace)
        if tablename in tablenames:
            return None
        elif test:
            return tablename

        indexes = []
        global_indexes = []
        hash_key = None

        if throughput is not None:
            table_throughput = Throughput(throughput['read'],
                                          throughput['write'])
        else:
            table_throughput = self.throughput

        hash_key = DynamoKey(self.hash_key.name,
                             data_type=self.hash_key.ddb_data_type)
        range_key = None
        if self.range_key is not None:
            range_key = DynamoKey(self.range_key.name,
                                  data_type=self.range_key.ddb_data_type)
        for field in six.itervalues(self.fields):
            if field.index:
                idx = field.get_ddb_index()
                indexes.append(idx)

        for gindex in self.global_indexes:
            index = gindex.get_ddb_index(self.fields)
            if throughput is not None and gindex.name in throughput:
                index.throughput = Throughput(**throughput[gindex.name])
            global_indexes.append(index)

        if not test:
            connection.create_table(tablename, hash_key, range_key,
                                    indexes, global_indexes, table_throughput)
            if wait:
                desc = connection.describe_table(tablename)
                while desc.status != 'ACTIVE':
                    time.sleep(1)
                    desc = connection.describe_table(tablename)

        return tablename

    def delete_dynamo_schema(self, connection, tablenames=None, test=False,
                             wait=False, namespace=()):
        """
        Drop all Dynamo tables for this model

        Parameters
        ----------
        connection : :class:`~dynamo3.DynamoDBConnection`
        tablenames : list, optional
            List of tables that already exist. Will call 'describe' if not
            provided.
        test : bool, optional
            If True, don't actually delete the table (default False)
        wait : bool, optional
            If True, block until table has been deleted (default False)
        namespace : str or tuple, optional
            The namespace of the table

        Returns
        -------
        table : str
            Table name that was deleted, or None if nothing deleted

        """
        if self.abstract:
            return None
        if tablenames is None:
            tablenames = set(connection.list_tables())

        tablename = self.ddb_tablename(namespace)
        if tablename in tablenames:
            if not test:
                connection.delete_table(tablename)
                if wait:
                    desc = connection.describe_table(tablename)
                    while desc is not None:
                        desc = connection.describe_table(tablename)
            return tablename
        return None



""" Query constraints """
import six

FILTER_ONLY = set(['contains', 'ncontains', 'null', 'in', 'ne'])


class Condition(object):

    """
    A constraint that will be applied to a query or scan

    Attributes
    ----------
    eq_fields : dict
        Mapping of field name to field value
    fields : dict
        Mapping of field name to (operator, value) tuples
    limit : int
        Maximum number of results
    index_name : str
        Name of index to use for a query

    """

    def __init__(self):
        self.eq_fields = {}
        self.fields = {}
        self.limit = None
        self.index_name = None

    def scan_kwargs(self):
        """ Get the kwargs for doing a table scan """
        kwargs = {}
        for key, val in six.iteritems(self.eq_fields):
            kwargs["%s__eq" % key] = val
        for key, (op, val) in six.iteritems(self.fields):
            kwargs["%s__%s" % (key, op)] = val
        if self.limit is not None:
            kwargs['limit'] = self.limit
        return kwargs

    def query_kwargs(self, model):
        """ Get the kwargs for doing a table query """
        if self.index_name is not None:
            ordering = model.meta_.get_ordering_from_index(self.index_name)
        else:
            queryable_keys = [k for k, (op, _) in six.iteritems(self.fields)
                              if op not in FILTER_ONLY]
            ordering = model.meta_.get_ordering_from_fields(
                self.eq_fields.keys(),
                queryable_keys,
            )

        if ordering is None:
            raise ValueError("Bad query arguments. You must provide a hash "
                             "key and may optionally constrain on exactly one "
                             "range key")
        kwargs = ordering.query_kwargs(self.eq_fields, self.fields)

        if self.limit is not None:
            kwargs['limit'] = self.limit
        return kwargs

    @classmethod
    def construct(cls, field, op, other):
        """
        Create a Condition on a field

        Parameters
        ----------
        field : str
            Name of the field to constrain
        op : str
            Operator, such as 'eq', 'lt', or 'contains'
        other : object
            The value to constrain the field with

        Returns
        -------
        condition : :class:`.Condition`

        """
        c = cls()
        if other is None:
            if op == 'eq':
                c.fields[field] = ('null', True)
            elif op == 'ne':
                c.fields[field] = ('null', False)
            else:
                raise ValueError("Cannot filter %s None" % op)
        elif op == 'eq':
            c.eq_fields[field] = other
        else:
            c.fields[field] = (op, other)
        return c

    @classmethod
    def construct_limit(cls, count):
        """
        Create a condition that will limit the results to a count

        Parameters
        ----------
        count : int

        Returns
        -------
        condition : :class:`.Condition`

        """
        c = cls()
        c.limit = count
        return c

    @classmethod
    def construct_index(cls, name):
        """
        Force the query to use a certain index

        Parameters
        ----------
        name : str

        Returns
        -------
        condition : :class:`.Condition`

        """
        c = cls()
        c.index_name = name
        return c

    def __and__(self, other):
        new_condition = Condition()
        new_condition.eq_fields.update(self.eq_fields)
        new_condition.fields.update(self.fields)
        new_condition.eq_fields.update(other.eq_fields)
        new_condition.fields.update(other.fields)
        if self.limit and other.limit:
            raise ValueError("Trying to combine two conditions with a "
                             "'limit' constraint!")
        new_condition.limit = self.limit or other.limit
        if self.index_name and other.index_name:
            raise ValueError("Trying to combine two conditions with an "
                             "'index' constraint!")
        new_condition.index_name = self.index_name or other.index_name
        return new_condition



""" Model code """
import six
import contextlib
import copy
import itertools

from dynamo3 import is_null
from .fields import Field, NUMBER
from .model_meta import ModelMetaclass, ModelMetadata, Ordering

# pylint: disable=E1002

SENTINEL = object()


class SetDelta(object):

    """
    Wrapper for an atomic change to a Dynamo set

    Used to track the changes when using :meth:`~.Model.add_` and
    :meth:`~.Model.remove_`

    """

    def __init__(self):
        self.action = None
        self.values = set()

    def merge(self, other):
        """
        Merge the delta with a set

        Parameters
        ----------
        other : set
            The original set to merge the changes with

        """
        if other is None:
            other = set()
        new = set()
        new.update(other)
        if self.action == 'ADD':
            new.update(self.values)
        elif other.issuperset(self.values):
            new.difference_update(self.values)
        else:
            raise KeyError("Cannot remove values that are not in the set!")

        return new

    def add(self, action, value):
        """
        Add another update to the delta

        Parameters
        ----------
        action : {'ADD', 'DELETE'}
        value : object
            The value to add or remove

        """
        if action not in ('ADD', 'DELETE'):
            raise ValueError("Invalid action '%s'" % action)
        if self.action is None:
            self.action = action

        if action == self.action:
            if isinstance(value, set):
                self.values.update(value)
            else:
                self.values.add(value)
        else:
            if not isinstance(value, set):
                value = set([value])
            if self.values.issuperset(value):
                self.values.difference_update(value)
            else:
                raise ValueError("Cannot ADD and REMOVE items from the same "
                                 "set in the same update")


class Model(six.with_metaclass(ModelMetaclass)):

    """
    Base class for all tube models

    For documentation on the metadata fields, check the attributes on the
    :class:`.ModelMetadata` class.

    Attributes
    ----------
    __metadata_class__ : class
        The class that is instantiated and set as ``meta_``
    __metadata__ : dict
        For details see :ref:`metadata`
    meta_ : :class:`~.ModelMetadata`
        The metadata for the model
    __engine__ : :class:`~flywheel.engine.Engine`
        Cached copy of the Engine that was used to save/load the model. This
        will be set after saving or loading a model.
    __dirty__ : set
        The set of all immutable fields that have been changed since the last
        save operation.
    __cache__ : dict
        The last seen value that was stored in the database. This is used to
        construct the ``expects`` dict when making updates that raise on
        conflict.
    __incrs__ : dict
        Mapping of fields to atomic add/delete operations for numbers and sets.

    """
    __metadata_class__ = ModelMetadata
    __metadata__ = {
        '_abstract': True,
    }
    meta_ = None
    __engine__ = None
    __dirty__ = None
    __cache__ = None
    __incrs__ = None
    # dict of all undeclared fields that have been set
    _overflow = None
    _persisted = False
    _loading = False

    def __init__(self, *args, **kwargs):  # pylint: disable=W0231
        if len(args) > 2 or (len(args) > 1 and self.meta_.range_key is None):
            raise TypeError("Too many positional arguments!")
        if len(args) > 0:
            setattr(self, self.meta_.hash_key.name, args[0])
        if len(args) > 1:
            setattr(self, self.meta_.range_key.name, args[1])
        for key, value in six.iteritems(kwargs):
            setattr(self, key, value)

    def refresh(self, consistent=False):
        """ Overwrite model data with freshest from database """
        if self.__engine__ is None:
            raise ValueError("Cannot sync: No DB connection")

        self.__engine__.refresh(self, consistent=consistent)

    def sync(self, raise_on_conflict=None, constraints=None):
        """ Sync model changes back to database """
        if self.__engine__ is None:
            raise ValueError("Cannot sync: No DB connection")

        self.__engine__.sync(self, raise_on_conflict=raise_on_conflict,
                             constraints=constraints)

    def delete(self, raise_on_conflict=None):
        """ Delete the model from the database """
        if self.__engine__ is None:
            raise ValueError("Cannot delete: No DB connection")
        self.__engine__.delete(self, raise_on_conflict=raise_on_conflict)

    @classmethod
    def __on_create__(cls):
        """ Called after class is constructed but before meta_ is set """
        pass

    @classmethod
    def __after_create__(cls):
        """ Called after class is constructed and after meta_ is set """
        pass

    def __new__(cls, *_, **__):
        """ Override __new__ to set default field values """
        obj = super(Model, cls).__new__(cls)
        mark_dirty = []
        with obj.loading_():
            for name, field in six.iteritems(cls.meta_.fields):
                if not field.composite:
                    setattr(obj, name, field.default)
                    if not is_null(field.default):
                        mark_dirty.append(name)
        obj.__dirty__.update(mark_dirty)
        obj._overflow = {}
        obj._persisted = False
        return obj

    def _is_field_primary(self, key):
        """ Check if a given field is part of the primary key """
        return ((self.meta_.hash_key.name in self.meta_.related_fields[key]) or
                (self.meta_.range_key is not None and
                 self.meta_.range_key.name in self.meta_.related_fields[key]))

    def __setattr__(self, name, value):
        if self.persisted_:
            if self._is_field_primary(name):
                if value != getattr(self, name):
                    raise AttributeError(
                        "Cannot change an item's primary key!")
                else:
                    return
        field = self.meta_.fields.get(name)
        if field is not None:
            # Ignore if trying to set a composite field
            if field.composite:
                return
            # Mutable fields check if they're dirty during sync()
            if field.is_mutable:
                return super(Model, self).__setattr__(name, field.coerce(value))

            # Don't mark the field dirty if the new and old values are the same
            oldv = getattr(self, name, SENTINEL)
            if not self._loading and oldv is not SENTINEL and oldv == value:
                return
            self.mark_dirty_(name)
            if (not self._loading and self.persisted_ and
                    name not in self.__cache__):
                for related in self.meta_.related_fields[name]:
                    cached_var = copy.copy(getattr(self, related))
                    self.__cache__[related] = cached_var
            return super(Model, self).__setattr__(name, field.coerce(value))
        elif name.startswith('_') or name.endswith('_'):
            # Don't interfere with non-Field private attrs
            return super(Model, self).__setattr__(name, value)
        else:
            self.mark_dirty_(name)
            if (not self._loading and self.persisted_ and name not in
                    self.__cache__):
                if Field.is_overflow_mutable(value):
                    self.__cache__[name] = copy.copy(self.get_(name))
                else:
                    self.__cache__[name] = self.get_(name)
            self._overflow[name] = value

    def __delattr__(self, name):
        field = self.meta_.fields.get(name)
        if field is not None:
            if not field.composite:
                setattr(self, name, None)
        elif name.startswith('_') or name.endswith('_'):
            # Don't interfere with non-Field private attrs
            super(Model, self).__delattr__(name)
            return
        else:
            setattr(self, name, None)

    def __getattribute__(self, name):
        if not name.startswith('__') and not name.endswith('_'):
            # Don't interfere with magic attrs or attrs ending in '_'
            field = self.meta_.fields.get(name)
            # Intercept getattribute to construct composite fields on the fly
            if field is not None and field.composite:
                return field.resolve(self)
        return super(Model, self).__getattribute__(name)

    def __getattr__(self, name):
        try:
            return self._overflow[name]
        except KeyError:
            raise AttributeError("%s not found" % name)

    def mark_dirty_(self, name):
        """ Mark that a field is dirty """
        if self._loading or self.__dirty__ is None:
            return
        if name in self.__incrs__:
            raise ValueError("Cannot increment field '%s' and set it in "
                             "the same update!" % name)
        if name in self.meta_.fields:
            self.__dirty__.update(self.meta_.related_fields[name])
            # Never mark the primary key as dirty
            if self.meta_.hash_key.name in self.__dirty__:
                self.__dirty__.remove(self.meta_.hash_key.name)
            if (self.meta_.range_key is not None and
                    self.meta_.range_key.name in self.__dirty__):
                self.__dirty__.remove(self.meta_.range_key.name)
        else:
            self.__dirty__.add(name)

    def get_(self, name, default=None):
        """ Dict-style getter for overflow attrs """
        return self._overflow.get(name, default)

    @property
    def hk_(self):
        """ The value of the hash key """
        return self.meta_.hk(self)

    @property
    def rk_(self):
        """ The value of the range key """
        return self.meta_.rk(self)

    @property
    def pk_dict_(self):
        """ The primary key dict, encoded for dynamo """
        return self.meta_.pk_dict(self, ddb_dump=True)

    def index_pk_dict_(self, index_name):
        """ The primary key dict for an index, encoded for dynamo """
        return self.meta_.index_pk_dict(self, ddb_dump=True)

    @property
    def pk_tuple_(self):
        """ The primary key dict, encoded for dynamo """
        return self.meta_.pk_tuple(self, ddb_dump=True)

    @property
    def persisted_(self):
        """ True if the model exists in DynamoDB, False otherwise """
        return self._persisted

    def keys_(self):
        """ All declared fields and any additional fields """
        return itertools.chain(self.meta_.fields.keys(), self._overflow.keys())

    def cached_(self, name, default=None):
        """ Get the cached (server) value of a field """
        if not self.persisted_:
            return default
        if name in self.__cache__:
            return self.__cache__[name]
        field = self.meta_.fields.get(name)
        # Need this redirection for Composite fields
        if field is not None:
            return field.get_cached_value(self)
        return getattr(self, name, default)

    def incr_(self, **kwargs):
        """ Atomically increment a number value """
        for key, val in six.iteritems(kwargs):
            if self._is_field_primary(key):
                raise AttributeError("Cannot increment an item's primary key!")

            field = self.meta_.fields.get(key)
            if field is not None:
                if field.ddb_data_type != NUMBER:
                    raise TypeError("Cannot increment non-number field '%s'" %
                                    key)
                if field.composite:
                    raise TypeError("Cannot increment composite field '%s'" %
                                    key)
            if key in self.__dirty__:
                raise ValueError("Cannot set field '%s' and increment it in "
                                 "the same update!" % key)
            self.__incrs__[key] = self.__incrs__.get(key, 0) + val
            if field is not None:
                self.__incrs__[key] = field.coerce(self.__incrs__[key], True)
                for name in self.meta_.related_fields[key]:
                    self.__cache__.setdefault(name, getattr(self, name))
                    if name != key:
                        self.__dirty__.add(name)
                self.__dict__[key] = self.cached_(key, 0) + self.__incrs__[key]
            else:
                self.__cache__.setdefault(key, getattr(self, key, 0))
                self._overflow[key] = self.cached_(
                    key, 0) + self.__incrs__[key]

    def add_(self, **kwargs):
        """ Atomically add to a set """
        self.mutate_('ADD', **kwargs)

    def remove_(self, **kwargs):
        """ Atomically remove from a set """
        self.mutate_('DELETE', **kwargs)

    def mutate_(self, action, **kwargs):
        """ Atomically mutate a set """
        for key, val in six.iteritems(kwargs):
            field = self.meta_.fields.get(key)
            if field is not None:
                if not field.is_set:
                    raise TypeError("Cannot mutate non-set field '%s'" %
                                    key)
                if field.composite:
                    raise TypeError("Cannot mutate composite field '%s'" %
                                    key)
            if key in self.__dirty__:
                raise ValueError("Cannot set field '%s' and mutate it in "
                                 "the same update!" % key)

            previous = self.__incrs__.get(key, SetDelta())
            previous.add(action, val)
            self.__incrs__[key] = previous
            if field is not None:
                for name in self.meta_.related_fields[key]:
                    self.__cache__.setdefault(name, getattr(self, name))
                    if name != key:
                        self.__dirty__.add(name)
                self.__dict__[key] = previous.merge(self.cached_(key))
            else:
                self.__cache__.setdefault(key, getattr(self, key, None))
                self._overflow[key] = previous.merge(self.cached_(key))

    def pre_save_(self, engine):
        """ Called before saving items """
        self.__engine__ = engine
        for field in six.itervalues(self.meta_.fields):
            field.validate(self)

    def post_save_(self):
        """ Called after item is saved to database """
        self._persisted = True
        self.__dirty__ = set()
        self.__incrs__ = {}
        self._reset_cache()

    def post_load_(self, engine):
        """ Called after model loaded from database """
        if engine is not None:
            self.__engine__ = engine
        self._persisted = True
        self.__dirty__ = set()
        self.__incrs__ = {}
        self._reset_cache()

    def _reset_cache(self):
        """ Reset the __cache__ to only track mutable fields """
        self.__cache__ = {}
        for name in self.keys_():
            field = self.meta_.fields.get(name)
            if field is None:
                value = self.get_(name)
                if Field.is_overflow_mutable(value):
                    self.__cache__[name] = copy.copy(value)
            elif not field.composite and field.is_mutable:
                self.__cache__[name] = copy.copy(getattr(self, name))

    @contextlib.contextmanager
    def loading_(self, engine=None):
        """ Context manager to speed up object load process """
        self._loading = True
        self._overflow = {}
        yield
        self._loading = False
        self.post_load_(engine)

    def ddb_dump_field_(self, name):
        """ Dump a field to a Dynamo-friendly value """
        val = getattr(self, name)
        if name in self.meta_.fields:
            return self.meta_.fields[name].ddb_dump(val)
        else:
            return Field.ddb_dump_overflow(val)

    def ddb_dump_(self):
        """ Return a dict for inserting into DynamoDB """
        data = {}
        for name in self.meta_.fields:
            data[name] = self.ddb_dump_field_(name)
        for name in self._overflow:
            data[name] = self.ddb_dump_field_(name)

        return data

    def set_ddb_val_(self, key, val):
        """ Decode and set a value retrieved from Dynamo """
        if key.startswith('_'):
            pass
        elif key in self.meta_.fields:
            setattr(self, key, self.meta_.fields[key].ddb_load(val))
        else:
            setattr(self, key, Field.ddb_load_overflow(val))

    @classmethod
    def ddb_load_(cls, engine, data):
        """ Load a model from DynamoDB data """
        obj = cls.__new__(cls)
        with obj.loading_(engine):
            for key, val in data.items():
                obj.set_ddb_val_(key, val)
        return obj

    def ddb_dump_cached_(self, name):
        """ Dump a cached field to a Dynamo-friendly value """
        val = self.cached_(name)
        if name in self.meta_.fields:
            return self.meta_.fields[name].ddb_dump(val)
        else:
            return Field.ddb_dump_overflow(val)

    def construct_ddb_expects_(self, fields=None):
        """ Construct a dynamo "expects" mapping based on the cached fields """
        if fields is None:
            fields = self.keys_()
        expect = {}
        for name in fields:
            val = self.ddb_dump_cached_(name)
            if val is None:
                expect[name + '__null'] = True
            else:
                expect[name + '__eq'] = val
        return expect

    @classmethod
    def field_(cls, name):
        """
        Construct a placeholder Field for an undeclared field

        This is used for creating scan filter constraints on fields that were
        not declared in the model

        """
        field = Field()
        field.name = name
        field.overflow = True
        return field

    def __json__(self, request=None):
        data = {}
        for name in self.meta_.fields:
            data[name] = getattr(self, name)
        for key, val in six.iteritems(self._overflow):
            data[key] = val
        return data

    def __hash__(self):
        return hash(self.hk_) + hash(self.rk_)

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.meta_.name == other.meta_.name and
                self.hk_ == other.hk_ and
                self.rk_ == other.rk_)

    def __ne__(self, other):
        return not self.__eq__(other)



""" Field type definitions """
import calendar
import datetime
import functools
import json
import six
from decimal import Decimal
from dynamo3 import (Binary, NUMBER, STRING, BINARY, NUMBER_SET, STRING_SET,
                     BINARY_SET, BOOL, MAP, LIST)
from dynamo3.types import float_to_decimal

from flywheel.compat import UnicodeMixin

ALL_TYPES = {}


def set_(data_type):
    """ Create an alias for a SetType that contains this data type """
    return frozenset([data_type])


def register_type(type_class, allow_in_set=True):
    """ Register a type class for use with Fields """
    ALL_TYPES[type_class.data_type] = type_class
    if allow_in_set:
        set_key = set_(type_class.data_type)
        ALL_TYPES[set_key] = SetType.bind(type_class.data_type)
    for alias in type_class.aliases:
        ALL_TYPES[alias] = type_class
        ALL_TYPES[set_(alias)] = SetType.bind(alias)


class TypeDefinition(UnicodeMixin):

    """
    Base class for all Field types

    Attributes
    ----------
    data_type : object
        The value you wish to pass in to Field as the data_type.
    aliases : list
        Other values that will reference this type if passed to Field
    ddb_data_type : {STRING, BINARY, NUMBER, STRING_SET, BINARY_SET, NUMBER_SET, BOOL, LIST, MAP}
        The DynamoDB data type that backs this type
    mutable : bool
        If True, flywheel will track updates to this field automatically when
        making calls to sync()
    allowed_filters : set
        The set of filters that can be used on this field type

    """
    data_type = None
    aliases = []
    ddb_data_type = None
    mutable = False

    def __init__(self):
        if self.ddb_data_type == NUMBER:
            self.allowed_filters = set(['eq', 'ne', 'lte', 'lt', 'gte', 'gt',
                                        'in', 'between'])
        elif self.ddb_data_type in (STRING, BINARY):
            self.allowed_filters = set(['eq', 'ne', 'lte', 'lt', 'gte', 'gt',
                                        'in', 'between', 'beginswith'])
        elif self.ddb_data_type in (NUMBER_SET, STRING_SET, BINARY_SET):
            self.allowed_filters = set(['contains', 'ncontains', 'in'])
        elif self.ddb_data_type in (MAP, BOOL):
            self.allowed_filters = set(['eq', 'ne'])
        elif self.ddb_data_type == LIST:
            self.allowed_filters = set(['eq', 'ne', 'contains', 'ncontains'])
        else:
            raise ValueError("Unknown dynamo data type '%s'" %
                             self.ddb_data_type)

    def coerce(self, value, force):
        """
        Check the type of a value and possible convert it

        Parameters
        ----------
        value : object
            The value to check
        force : bool
            If True, always attempt to convert a bad type to the correct type

        Returns
        -------
        value : object
            A variable of the correct type

        Raises
        ------
        exc : TypeError or ValueError
            If the value is the incorrect type and could not be converted

        """

        if not isinstance(value, self.data_type):
            raise TypeError()
        return value

    def ddb_dump_inner(self, value):
        """
        If this is a set type, dump a value to the type contained in the set

        """
        return value

    def ddb_dump(self, value):
        """ Dump a value to a form that can be stored in DynamoDB """
        return value

    def ddb_load(self, value):
        """ Turn a value into this type from a DynamoDB value """
        return value

    def _attempt_coerce_json(self, value, obj_type):
        """
        If a value was previously stored as json, attempt to load it as an
        obj_type object
        """
        if isinstance(value, six.text_type):
            orig_value = value
            try:
                value = json.loads(orig_value)
                if not isinstance(value, obj_type):
                    value = orig_value
            except Exception as e:
                raise TypeError(e)
        return value

    def __repr__(self):
        return 'TypeDefinition(%s)' % self

    def __unicode__(self):
        return six.text_type(self.data_type)


class SetType(TypeDefinition):

    """ Set types """
    data_type = set
    mutable = True

    def __init__(self, item_type=None, type_class=None):
        self.item_type = item_type
        set_map = {
            STRING: STRING_SET,
            NUMBER: NUMBER_SET,
            BINARY: BINARY_SET,
        }
        if item_type is not None and type_class is None:
            self.item_field = ALL_TYPES[item_type]()
            self.ddb_data_type = set_map[self.item_field.ddb_data_type]
        elif type_class is not None:
            self.item_field = type_class()
            self.ddb_data_type = set_map[self.item_field.ddb_data_type]
        else:
            self.item_field = None
            self.ddb_data_type = STRING_SET
        super(SetType, self).__init__()

    def coerce(self, value, force):
        if not isinstance(value, set):
            if force:
                value = set(value)
            else:
                raise TypeError()
        if self.item_field is not None:
            converted_values = set()
            for item in value:
                converted_values.add(self.item_field.coerce(item, force))
            return converted_values
        return value

    def ddb_dump_inner(self, value):
        """ We need to expose this for 'contains' and 'ncontains' """
        if self.item_field is None:
            return value
        return self.item_field.ddb_dump(value)

    def ddb_dump(self, value):
        if self.item_field is None:
            return value
        return set([self.ddb_dump_inner(v) for v in value])

    def ddb_load(self, value):
        if self.item_field is None:
            return value
        return set([self.item_field.ddb_load(v) for v in value])

    def __unicode__(self):
        if self.item_type is None:
            return super(SetType, self).__unicode__()
        else:
            return six.text_type(set([self.item_type]))

    @classmethod
    def bind(cls, item_type):
        """ Create a set factory that will contain a specific data type """
        return functools.partial(cls, item_type)

register_type(SetType, allow_in_set=False)
ALL_TYPES[STRING_SET] = SetType.bind(STRING)
ALL_TYPES[BINARY_SET] = SetType.bind(BINARY)
ALL_TYPES[NUMBER_SET] = SetType.bind(NUMBER)


class NumberType(TypeDefinition):

    """ Any kind of numerical value """
    data_type = NUMBER
    ddb_data_type = NUMBER

    def coerce(self, value, force):
        if not (isinstance(value, float) or isinstance(value, Decimal) or
                isinstance(value, six.integer_types)):
            if force:
                try:
                    return int(value)
                except ValueError:
                    return float(value)
            else:
                raise TypeError()
        return value

    def ddb_load(self, value):
        if isinstance(value, Decimal):
            if value % 1 == 0:
                return int(value)
            else:
                return float(value)
        return value

register_type(NumberType)


class FloatType(TypeDefinition):

    """ Float values """
    data_type = float
    ddb_data_type = NUMBER

    def coerce(self, value, force):
        if not isinstance(value, float):
            # Auto-convert ints, longs, and Decimals
            if (isinstance(value, six.integer_types) or
                    isinstance(value, Decimal)):
                return float(value)
            elif force:
                return float(value)
            else:
                raise TypeError()
        return value

    def ddb_load(self, value):
        return float(value)

register_type(FloatType)


class IntType(TypeDefinition):

    """ Integer values (includes longs) """
    data_type = int
    aliases = list(six.integer_types)
    ddb_data_type = NUMBER

    def coerce(self, value, force):
        if not isinstance(value, six.integer_types):
            if force:
                new_val = int(value)
                if isinstance(value, float) or isinstance(value, Decimal):
                    if new_val != value:
                        raise ValueError("Refusing to convert "
                                         "%r to int! Results in data loss!"
                                         % repr(value))
                return new_val
            else:
                raise TypeError()
        return value

    def ddb_load(self, value):
        return int(value)

register_type(IntType)


class DecimalType(TypeDefinition):

    """
    Numerical values that use Decimal in the application layer.

    This should be used if you want to work with floats but need the additional
    precision of the Decimal type.

    """

    data_type = Decimal
    ddb_data_type = NUMBER

    def coerce(self, value, force):
        if not isinstance(value, Decimal):
            if force:
                # Python 2.6 can't convert directly from float to Decimal
                if (isinstance(value, float) and
                        not hasattr(Decimal, 'from_float')):
                    return float_to_decimal(value)
                return Decimal(value)
            else:
                raise TypeError()
        return value

register_type(DecimalType)


class BoolType(TypeDefinition):

    """ Boolean type """

    data_type = bool
    ddb_data_type = BOOL

    def __init__(self):
        super(BoolType, self).__init__()

    def coerce(self, value, force):
        value = self._attempt_coerce_json(value, bool)
        if not isinstance(value, bool):
            if force:
                return bool(value)
            else:
                raise TypeError()
        return value

register_type(BoolType)


class StringType(TypeDefinition):

    """ String values, stored as unicode """
    data_type = six.text_type
    aliases = [STRING]
    ddb_data_type = STRING

    def coerce(self, value, force):
        if not isinstance(value, six.text_type):
            # Silently convert str to unicode using utf-8
            if isinstance(value, six.binary_type):
                return value.decode('utf-8')
            if force:
                return six.text_type(value)
            else:
                raise TypeError()
        return value

register_type(StringType)


class BinaryType(TypeDefinition):

    """ Binary strings, stored as a str/bytes """
    data_type = six.binary_type
    aliases = [BINARY, Binary]
    ddb_data_type = BINARY

    def coerce(self, value, force):
        if not isinstance(value, six.binary_type):
            # Silently convert unicode to str using utf-8
            if isinstance(value, six.text_type):
                return value.encode('utf-8')
            if force:
                return six.binary_type(value)
            else:
                raise TypeError()
        return value

    def ddb_dump(self, value):
        return Binary(value)

    def ddb_load(self, value):
        return value.value

register_type(BinaryType)


class DictType(TypeDefinition):

    """ Dict type, stored as a map """
    data_type = dict
    ddb_data_type = MAP
    mutable = True

    def __init__(self):
        super(DictType, self).__init__()

    def coerce(self, value, force):
        value = self._attempt_coerce_json(value, dict)
        if not isinstance(value, dict):
            if force:
                return dict(value)
            else:
                raise TypeError()
        return value

register_type(DictType)


class ListType(TypeDefinition):

    """ List type """
    data_type = list
    ddb_data_type = LIST
    mutable = True

    def __init__(self):
        super(ListType, self).__init__()

    def coerce(self, value, force):
        value = self._attempt_coerce_json(value, list)
        if not isinstance(value, list):
            if force:
                return list(value)
            else:
                raise TypeError()
        return value

register_type(ListType)


ZERO = datetime.timedelta(0)

# https://docs.python.org/2/library/datetime.html#datetime.tzinfo.fromutc


class UTCTimezone(datetime.tzinfo):

    """ UTC """

    def utcoffset(self, dt):
        return ZERO

    def tzname(self, dt):
        return 'UTC'

    def dst(self, dt):
        return ZERO

UTC = UTCTimezone()


class DateTimeType(TypeDefinition):

    """
    Datetimes, stored as a unix timestamp

    Parameters
    ----------
    naive : bool, optional
        If True, will load values from Dynamo with no timezone. If False, will
        add a UTC timezone. (Default False).

    Notes
    -----
    If you want to use naive datetimes, you will need to reference the type
    class directly instead of going through an alias. For example:

    .. code-block:: python

        from flywheel.fields.types import DateTimeType

        field = Field(data_type=DateTimeType(naive=True))

    """
    data_type = datetime.datetime
    ddb_data_type = NUMBER

    def __init__(self, naive=False):
        super(DateTimeType, self).__init__()
        self.naive = naive

    def ddb_dump(self, value):
        seconds = calendar.timegm(value.utctimetuple())
        milliseconds = value.strftime('%f')
        return Decimal("%d.%s" % (seconds, milliseconds))

    def ddb_load(self, value):
        microseconds = int(1000000 * (value - int(value)))
        dt = datetime.datetime.utcfromtimestamp(value) \
            .replace(microsecond=microseconds)
        if self.naive:
            return dt
        else:
            return dt.replace(tzinfo=UTC)

register_type(DateTimeType)


class DateType(TypeDefinition):

    """ Dates, stored as timestamps """
    data_type = datetime.date
    ddb_data_type = NUMBER

    def ddb_dump(self, value):
        return calendar.timegm(value.timetuple())

    def ddb_load(self, value):
        return datetime.datetime.utcfromtimestamp(value).date()


register_type(DateType)



""" Setup file """
import sys

from setuptools import setup, find_packages

import os
import re


HERE = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(HERE, 'README.rst')).read()
CHANGES = open(os.path.join(HERE, 'CHANGES.rst')).read()
# Remove custom RST extensions for pypi
CHANGES = re.sub(r'\(\s*:(issue|pr|sha):.*?\)', '', CHANGES)

REQUIREMENTS = [
    'dynamo3>=0.4.6',
    'six'
]

TEST_REQUIREMENTS = [
    'nose',
    'mock',
]

if sys.version_info[:2] < (2, 7):
    TEST_REQUIREMENTS.extend(['unittest2'])

if __name__ == "__main__":
    setup(
        name='flywheel',
        version='0.4.7',
        description="SQLAlchemy-style ORM for Amazon's DynamoDB",
        long_description=README + '\n\n' + CHANGES,
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.6',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Topic :: Database',
        ],
        author='Steven Arcangeli',
        author_email='stevearc@stevearc.com',
        url='http://flywheel.readthedocs.org/',
        license='MIT',
        keywords='aws dynamo dynamodb orm odm',
        platforms='any',
        include_package_data=True,
        packages=find_packages(exclude=('tests',)),
        install_requires=REQUIREMENTS,
        tests_require=REQUIREMENTS + TEST_REQUIREMENTS,
    )



""" Field declarations for models """
import copy
import six
import inspect
import json
from dynamo3 import (DynamoKey, LocalIndex, NUMBER, STRING, BINARY, NUMBER_SET,
                     STRING_SET, BINARY_SET, is_null)
from decimal import Decimal

from .conditions import Condition
from .indexes import GlobalIndex
from .types import TypeDefinition, ALL_TYPES, set_

NO_ARG = object()


class Field(object):

    """
    Declarative way to specify model fields

    Parameters
    ----------
    hash_key : bool, optional
        This key is a DynamoDB hash key (default False)
    range_key : bool, optional
        This key is a DynamoDB range key (default False)
    index : str, optional
        If present, create a local secondary index on this field with this as
        the name.
    data_type : object, optional
        The field data type. You may use int, unicode, set, etc. or you may
        pass in an instance of :class:`~flywheel.fields.types.TypeDefinition`
        (default unicode)
    coerce : bool, optional
        Attempt to coerce the value if it's the incorrect type (default False)
    check : callable or list, optional
        A function that takes the value and returns True if the value is valid.
        May also be a list of such functions. (default None)
    nullable : bool, optional
        If false, will add a check (above) to ensure the value is not null
        (default True).
    default : object, optional
        The default value for this field that will be set when creating a model
        (default None, except for ``set`` data types which default to set())

    Attributes
    ----------
    name : str
        The name of the attribute on the model
    model : class
        The :class:`.Model` this field is attached to
    composite : bool
        True if this is a composite field

    Notes
    -----
    ::

        Field(index='my-index')

    Is shorthand for::

        Field().all_index('my-index')

    """

    def __init__(self, hash_key=False, range_key=False, index=None,
                 data_type=six.text_type, coerce=False, check=None,
                 nullable=True, default=NO_ARG):
        if hash_key and range_key:
            raise ValueError("hash_key and range_key are mutually exclusive!")
        self.name = None
        self.model = None
        self.composite = False
        self.overflow = False
        if isinstance(data_type, TypeDefinition):
            self.data_type = data_type
        elif (inspect.isclass(data_type) and
              issubclass(data_type, TypeDefinition)):
            self.data_type = data_type()
        else:
            type_factory = ALL_TYPES.get(data_type)
            if type_factory is None:
                raise TypeError("Unrecognized data_type '%s'" % data_type)
            self.data_type = type_factory()
        self._coerce = coerce
        self.check = []
        if check is not None:
            if hasattr(check, '__iter__'):
                self.check = list(check)
            else:
                self.check = [check]
        if not nullable:
            self.check.append(lambda x: not is_null(x))
        self.hash_key = hash_key
        self.range_key = range_key
        self.subfields = []
        self.index = False
        self.index_name = None
        self._ddb_index = None
        self._ddb_index_kwargs = None
        if default is NO_ARG:
            if self.is_set:
                self._default = set()
            else:
                self._default = None
        else:
            self._default = default
        if index:
            self.all_index(index)

    @property
    def default(self):
        """ Get a shallow copy of the default value """
        return copy.copy(self._default)

    def get_ddb_index(self):
        """ Construct a dynamo local index object """
        range_key = DynamoKey(self.name, data_type=self.ddb_data_type)
        return self._ddb_index(self.index_name, range_key,
                               **self._ddb_index_kwargs)

    def _set_ddb_index(self, name, factory, **kwargs):
        """ Set the type of index """
        if self.hash_key or self.range_key:
            raise ValueError("Cannot index the hash or range key!")
        if self.index:
            raise ValueError("Index is already set!")
        self.index_name = name
        self._ddb_index = factory
        self._ddb_index_kwargs = kwargs
        self.index = True

    def validate(self, obj):
        """
        Run the validation checks for this field on a model object.

        Parameters
        ----------
        obj : :class:`~flywheel.models.Model`

        Raises
        ------
        err : :class:`ValueError`
            Raised if any of the checks fail.

        """
        if self.check is None:
            return
        val = self.resolve(obj)
        for check in self.check:
            if not check(val):
                raise ValueError("Validation check on field %s failed for "
                                 "value %r" % (self.name, val))

    def all_index(self, name):
        """
        Index this field and project all attributes

        Parameters
        ----------
        name : str
            The name of the index

        """
        self._set_ddb_index(name, LocalIndex.all)
        return self

    def keys_index(self, name):
        """
        Index this field and project all key attributes

        Parameters
        ----------
        name : str
            The name of the index

        """
        self._set_ddb_index(name, LocalIndex.keys)
        return self

    def include_index(self, name, includes=None):
        """
        Index this field and project selected attributes

        Parameters
        ----------
        name : str
            The name of the index
        includes : list, optional
            List of non-key attributes to project into this index

        """
        includes = includes or []
        self._set_ddb_index(name, LocalIndex.include, includes=includes)
        return self

    def coerce(self, value, force_coerce=None):
        """ Coerce the value to the field's data type """
        if value is None:
            return value
        if force_coerce is None:
            force_coerce = self._coerce
        try:
            return self.data_type.coerce(value, force_coerce)
        except (TypeError, ValueError) as e:
            if e.args:
                raise
            raise TypeError("Field '%s' must be %s! '%s'" %
                            (self.name, self.data_type, repr(value)))

    @property
    def is_mutable(self):
        """ Return True if the data type is mutable """
        return self.data_type.mutable

    @property
    def is_set(self):
        """ Return True if data type is a set """
        return self.ddb_data_type in (STRING_SET, NUMBER_SET, BINARY_SET)

    def ddb_dump(self, value):
        """ Dump a value to its Dynamo format """
        if value is None:
            return None
        return self.data_type.ddb_dump(value)

    def ddb_dump_for_query(self, value):
        """ Dump a value to format for use in a Dynamo query """
        if value is None:
            return None
        if self.overflow:
            return self.ddb_dump_overflow(value)
        value = self.coerce(value, force_coerce=True)
        return self.ddb_dump(value)

    def ddb_load(self, val):
        """ Decode a value retrieved from Dynamo """
        return self.data_type.ddb_load(val)

    @classmethod
    def ddb_dump_overflow(cls, val):
        """ Dump an overflow value to its Dynamo format """
        if val is None:
            return None
        elif isinstance(val, six.integer_types) or isinstance(val, float):
            return val
        elif isinstance(val, set):
            return val
        else:
            return json.dumps(val)

    @classmethod
    def ddb_load_overflow(cls, val):
        """ Decode a value of an overflow field """
        if (isinstance(val, Decimal) or isinstance(val, float) or
                isinstance(val, six.integer_types)):
            if val % 1 == 0:
                return int(val)
            return float(val)
        elif isinstance(val, set):
            return val
        else:
            return json.loads(val)

    @classmethod
    def is_overflow_mutable(cls, val):
        """ Check if an overflow field is mutable """
        if val is None:
            return False
        val_type = type(val)
        if val_type in ALL_TYPES:
            return ALL_TYPES[val_type].mutable
        return True

    @property
    def ddb_data_type(self):
        """ Get the native DynamoDB data type """
        return self.data_type.ddb_data_type

    def can_resolve(self, fields):
        """
        Check if the provided fields are enough to fully resolve this field

        Parameters
        ----------
        fields : list or set

        Returns
        -------
        needed : set
            Set of the subfields needed to resolve this field. If empty, then
            it cannot be resolved.

        """
        needed = set()
        # If this field name is found in the fields, that is all that's needed
        if self.name in fields:
            needed.add(self.name)
        elif self.subfields:
            # Otherwise, we need ALL of the subfields (only non-empty for
            # Composites) to recursively be satisfied
            for field in self.subfields:
                resolve = self.model.meta_.fields[field].can_resolve(fields)
                if not resolve:
                    return set()
                needed.update(resolve)
        return needed

    def resolve(self, obj=None, scope=None):
        """ Resolve a field value from an object or scope dict """
        if obj is not None:
            return getattr(obj, self.name)
        else:
            return scope[self.name]

    def get_cached_value(self, obj):
        """ Get the cached value of a field before any local modifications """
        if self.name in obj.__cache__:
            return obj.__cache__[self.name]
        else:
            return self.resolve(obj)

    def _make_condition(self, filter, other):
        """
        Construct a query condition for a filter on a value

        Parameters
        ----------
        filter : str
            The name of the filter (e.g. 'eq' or 'lte')
        other : object
            The value to filter on

        """
        other = self.ddb_dump_for_query(other)
        if other is None and filter in ('ne', 'eq'):
            # Don't bother checking allowed filters because this turns into the
            # "exists" filter, which can be done on anything
            pass
        elif self.overflow:
            # Don't bother checking validity if this is an overflow field
            pass
        elif filter not in self.data_type.allowed_filters:
            raise TypeError("Cannot use '%s' filter on '%s' field" %
                            (filter, self.data_type))
        return Condition.construct(self.name, filter, other)

    def __eq__(self, other):
        return self._make_condition('eq', other)

    def __ne__(self, other):
        return self._make_condition('ne', other)

    def __lt__(self, other):
        return self._make_condition('lt', other)

    def __le__(self, other):
        return self._make_condition('lte', other)

    def __gt__(self, other):
        return self._make_condition('gt', other)

    def __ge__(self, other):
        return self._make_condition('gte', other)

    def _make_contains_condition(self, filter, other):
        """ Construct a query condition for 'contains' or 'ncontains' """
        if not self.overflow and filter not in self.data_type.allowed_filters:
            raise TypeError("Cannot use '%s' filter on '%s' field" %
                            (filter, self.data_type))
        return Condition.construct(self.name, filter,
                                   self.data_type.ddb_dump_inner(other))

    def contains_(self, other):
        """ Create a query condition that this field must contain a value """
        return self._make_contains_condition('contains', other)

    def ncontains_(self, other):
        """ Create a query condition that this field cannot contain a value """
        return self._make_contains_condition('ncontains', other)

    def in_(self, other):
        """
        Create a query condition that this field must be within a set of values

        """
        if self.overflow:
            other = set([self.ddb_dump_overflow(val) for val in other])
        elif 'in' not in self.data_type.allowed_filters:
            raise TypeError("Cannot use '%s' filter on '%s' field" %
                            (filter, self.data_type))
        else:
            other = set([self.ddb_dump_for_query(val) for val in other])
        return Condition.construct(self.name, 'in', other)

    def beginswith_(self, other):
        """
        Create a query condition that this field must begin with a string

        """
        if (not self.overflow and
                'beginswith' not in self.data_type.allowed_filters):
            raise TypeError("Cannot use 'beginswith' filter on '%s' field" %
                            self.data_type)
        if self.overflow:
            # Since strings are dumped to json in overflow fields, we should
            # prefix it with a double quote
            other = '"' + other
        else:
            other = self.ddb_dump_for_query(other)
        return Condition.construct(self.name, 'beginswith', other)

    def between_(self, low, high):
        """
        Create a query condition that this field must be between two values
        (inclusive)

        """
        if (not self.overflow and
                'between' not in self.data_type.allowed_filters):
            raise TypeError("Cannot use 'between' filter on %s field" %
                            self.data_type)
        low = self.ddb_dump_for_query(low)
        high = self.ddb_dump_for_query(high)
        return Condition.construct(self.name, 'between', (low, high))

    def betwixt_(self, low, high):
        """ Poetic version of :meth:`~.between_` """
        return self.between_(low, high)


class Composite(Field):

    """
    A field that is composed of multiple other fields

    Parameters
    ----------
    *fields : list
        List of names of fields that compose this composite field
    hash_key : bool, optional
        This key is a DynamoDB hash key (default False)
    range_key : bool, optional
        This key is a DynamoDB range key (default False)
    index : str, optional
        If present, create a local secondary index on this field with this as
        the name.
    data_type : str, optional
        The dynamo data type. Valid values are (NUMBER, STRING, BINARY,
        NUMBER_SET, STRING_SET, BINARY_SET, dict, list, bool, str, unicode,
        int, float, set, datetime, date, Decimal) (default unicode)
    coerce : bool, optional
        Attempt to coerce the value if it's the incorrect type (default False)
    check : callable, optional
        A function that takes the value and returns True if the value is valid
        (default None)
    merge : callable, optional
        The function that merges the subfields together. By default it simply
        joins them with a ':'.

    """

    def __init__(self, *args, **kwargs):
        self.merge = kwargs.pop('merge', None)
        if self.merge is None:
            self.merge = lambda *args: ':'.join(args)
        unrecognized = (set(kwargs.keys()) -
                        set(['range_key', 'index', 'hash_key', 'data_type',
                             'check', 'coerce']))
        if unrecognized:
            raise TypeError("Unrecognized keyword args: %s" % unrecognized)
        if len(args) < 1:
            raise TypeError("Composite must depend on at least one field")
        super(Composite, self).__init__(**kwargs)
        self.composite = True
        self.subfields = args

    def __contains__(self, key):
        return key in self.subfields

    def resolve(self, obj=None, scope=None):
        """ Resolve a field value from an object or scope dict """
        if scope is not None and self.name in scope:
            return super(Composite, self).resolve(obj, scope)
        args = [self.model.meta_.fields[f].resolve(obj, scope) for f in
                self.subfields]
        return self.coerce(self.merge(*args))

    def get_cached_value(self, obj):
        args = [self.model.meta_.fields[f].get_cached_value(obj) for f in
                self.subfields]
        return self.coerce(self.merge(*args))



""" Query and Scan builders """
import six

from .fields import Condition


class EntityNotFoundException(ValueError):

    """ Raised when results are expected and not found. """


class DuplicateEntityException(ValueError):

    """ Raised when too many results are found. """


class Query(object):

    """
    An object used to query dynamo tables

    See the :class:`.Engine` for query examples

    Parameters
    ----------
    engine : :class:`.Engine`
    model : class
        Subclass of :class:`~flywheel.models.Model`

    """

    def __init__(self, engine, model):
        self.engine = engine
        self.model = model
        self.condition = Condition()

    @property
    def dynamo(self):
        """ Shortcut to access DynamoDBConnection """
        return self.engine.dynamo

    @property
    def tablename(self):
        """ Shortcut to access dynamo table name """
        return self.model.meta_.ddb_tablename(self.engine.namespace)

    def gen(self, desc=False, consistent=False, attributes=None,
            filter_or=False, exclusive_start_key=None):
        """
        Return the query results as a generator

        Parameters
        ----------
        desc : bool, optional
            Return results in descending order (default False)
        consistent : bool, optional
            Force a consistent read of the data (default False)
        attributes : list, optional
            List of fields to retrieve from dynamo. If supplied, gen() will
            iterate over dicts instead of model objects.
        filter_or : bool, optional
            If True, multiple filter() constraints will be joined with an OR
            (default AND).
        exclusive_start_key : dict, optional
            The ExclusiveStartKey to resume a previous query

        Returns
        -------
        results : generator

        """
        kwargs = self.condition.query_kwargs(self.model)
        if attributes is not None:
            kwargs['attributes'] = attributes
        results = self.dynamo.query(
            self.tablename, desc=desc, consistent=consistent,
            filter_or=filter_or, exclusive_start_key=exclusive_start_key,
            **kwargs)
        for result in results:
            if attributes is not None:
                yield result
            else:
                yield self.model.ddb_load_(self.engine, result)

    def __iter__(self):
        return self.gen()

    def all(self, desc=False, consistent=False, attributes=None,
            filter_or=False, exclusive_start_key=None):
        """
        Return the query results as a list

        Parameters
        ----------
        desc : bool, optional
            Return results in descending order (default False)
        consistent : bool, optional
            Force a consistent read of the data (default False)
        attributes : list, optional
            List of fields to retrieve from dynamo. If supplied, returns dicts
            instead of model objects.
        filter_or : bool, optional
            If True, multiple filter() constraints will be joined with an OR
            (default AND).
        exclusive_start_key : dict, optional
            The ExclusiveStartKey to resume a previous query

        Returns
        -------
        results : list

        """
        return list(self.gen(desc=desc, consistent=consistent,
                             attributes=attributes, filter_or=filter_or,
                             exclusive_start_key=exclusive_start_key))

    def count(self, filter_or=False):
        """
        Find the number of elements the match this query

        Parameters
        ----------
        filter_or : bool, optional
            If True, multiple filter() constraints will be joined with an OR
            (default AND).

        Returns
        -------
        count : int

        """
        kwargs = self.condition.query_kwargs(self.model)
        return self.dynamo.query(self.tablename, count=True,
                                 filter_or=filter_or, **kwargs)

    def first(self, desc=False, consistent=False, attributes=None,
              filter_or=False):
        """
        Return the first result of the query, or None if no results

        Parameters
        ----------
        desc : bool, optional
            Return results in descending order (default False)
        consistent : bool, optional
            Force a consistent read of the data (default False)
        attributes : list, optional
            List of fields to retrieve from dynamo. If supplied, returns dicts
            instead of model objects.
        filter_or : bool, optional
            If True, multiple filter() constraints will be joined with an OR
            (default AND).

        Returns
        -------
        result : :class:`~flywheel.models.Model` or None

        """
        self.limit(1)
        for result in self.gen(desc=desc, consistent=consistent,
                               attributes=attributes, filter_or=filter_or):
            return result
        return None

    def one(self, consistent=False, attributes=None, filter_or=False):
        """
        Return the result of the query. If there is not exactly one result,
        raise a ValueError

        Parameters
        ----------
        consistent : bool, optional
            Force a consistent read of the data (default False)
        attributes : list, optional
            List of fields to retrieve from dynamo. If supplied, returns dicts
            instead of model objects.
        filter_or : bool, optional
            If True, multiple filter() constraints will be joined with an OR
            (default AND).

        Returns
        -------
        result : :class:`~flywheel.models.Model`

        Raises
        ------
        e1 : :class:`~.EntityNotFoundException`
            If no entity is found. Subclasses :class:`~ValueError`.
        e2 : :class:`~.DuplicateEntityException`
            If more than one entity is found. Subclasses :class:`~ValueError`.

        """
        self.limit(2)
        items = self.all(consistent=consistent, attributes=attributes,
                         filter_or=filter_or)
        if len(items) > 1:
            raise DuplicateEntityException("More than one result!")
        elif len(items) == 0:
            raise EntityNotFoundException("Expected one result!")
        return items[0]

    def limit(self, count):
        """ Limit the number of query results """
        self.condition &= Condition.construct_limit(count)
        return self

    def index(self, name):
        """ Use a specific local or global index for the query """
        self.condition &= Condition.construct_index(name)
        return self

    def delete(self, filter_or=False):
        """
        Delete all items that match the query

        Parameters
        ----------
        filter_or : bool, optional
            If True, multiple filter() constraints will be joined with an OR
            (default AND).

        """
        meta = self.model.meta_
        attrs = [meta.hash_key.name]
        if meta.range_key is not None:
            attrs.append(meta.range_key.name)
        results = self.gen(attributes=attrs, filter_or=filter_or)
        return self.engine.delete_key(self.model, results)

    def filter(self, *conditions, **kwargs):
        """
        Add a Condition to constrain the query

        Notes
        -----
        The conditions may be passed in as positional arguments:

        .. code-block:: python

            engine.query(User).filter(User.id == 12345)

        Or they may be passed in as keyword arguments:

        .. code-block:: python

            engine.query(User).filter(firstname='Monty', lastname='Python')

        The limitations of the keyword method is that you may only create
        equality conditions. You may use both types in a single filter:

        .. code-block:: python

            engine.query(User).filter(User.num_friends > 10, name='Monty')

        """
        for condition in conditions:
            self.condition &= condition
        for key, val in six.iteritems(kwargs):
            field = self.model.meta_.fields.get(key)
            if field is not None:
                self.condition &= (field == val)
            else:
                self.condition &= (self.model.field_(key) == val)

        return self


class Scan(Query):

    """
    An object used to scan dynamo tables

    scans are like Queries except they don't use indexes. This means they
    iterate over all data in the table and are SLOW

    Parameters
    ----------
    engine : :class:`.Engine`
    model : class
        Subclass of :class:`~flywheel.models.Model`

    """

    def gen(self, attributes=None, desc=False, consistent=False,
            filter_or=False, exclusive_start_key=None):
        if desc:
            raise ValueError("Cannot order scan() results")
        if consistent:
            raise ValueError("Cannot force consistent read on scan()")
        kwargs = self.condition.scan_kwargs()

        if attributes is not None:
            kwargs['attributes'] = attributes
        results = self.dynamo.scan(
            self.tablename, filter_or=filter_or,
            exclusive_start_key=exclusive_start_key, **kwargs)
        for result in results:
            if attributes is not None:
                yield result
            else:
                yield self.model.ddb_load_(self.engine, result)

    def count(self, filter_or=False):
        kwargs = self.condition.scan_kwargs()
        return self.dynamo.scan(self.tablename, count=True, filter_or=False,
                                **kwargs)

    def index(self, name):
        raise TypeError("Scan cannot use an index!")



""" Utilities for Python 2/3 compatibility """
import six


class UnicodeMixin(object):

    """ Mixin that handles __str__ and __bytes__. Just define __unicode__.  """
    if six.PY3:  # pragma: no cover
        def __str__(self):
            return self.__unicode__()

        def __bytes__(self):
            return self.__unicode__().encode('utf-8')
    else:
        def __str__(self):
            return self.__unicode__().encode('utf-8')
